from .poly import *
from diogi.functions import (
    always_a_list,
    get_if_exists, 
    none_if_empty, 
    set_if_not_none
)

def action(action: Action) -> dict:
        action_type = 'safe'
        return {
            'name': action.name, 
            'type': action_type, 
            'url': action.url
        } 

class Wstl:
    MIME_TYPE = 'application/prs.wstl+json'

    def format(poly: Poly) -> dict:
        data = {}
        set_if_not_none(data, poly.title, 'title')

        actions = list(map(action, poly.actions))
        set_if_not_none(data, none_if_empty(actions), 'actions') 

        set_if_not_none(data, none_if_empty(poly.data), 'data')
        
        return {
            'wstl': data
        }

    def parse(wstl: dict) -> Poly: 
        wstl = wstl['wstl']
        poly = Poly()
        poly.set_title(get_if_exists(wstl, 'title'))

        actions = always_a_list(get_if_exists(wstl, 'actions'))
        for a in actions: 
            poly.actions.append(Action(
                get_if_exists(a, 'name'),
                get_if_exists(a, 'url')
                ))

        data = always_a_list(get_if_exists(wstl, 'data'))
        for d in data: 
            poly.data.append(d)

        return poly


import requests
from alps import Alps, Unsafe
from .poly import Poly 
from .wstl import Wstl 
from .hal import Hal 
from diogi.functions import first_or_default


def get_alps():
    pass


class RestClient: 
    def __init__(self):
        self.formatter = None
        self.format = 'json'
        self.poly = Poly()
        self._accepted_types = ''
        self._alps = None 

    def connect(self, url: str) -> Poly:
        headers = {
            'Accept': f'{Wstl.MIME_TYPE}, {Hal.MIME_TYPE}, application/json'
        }
        
        response = requests.get(url, headers=headers) 
        self._accepted_types = response.headers['Content-Type']
        data = response.json()

        if 'wstl' in data.keys(): 
            self.format = 'wstl'
            self.formatter = Wstl
            self.poly = Wstl.parse(response.json())
        else: 
            self.poly = Poly()
            self.poly.add_data_item(data)
        
        self._detect_and_set_alps()

        return self.poly

    def get(self, action: str, resolver: callable = None) -> dict:
        params = self._resolve_params(action, resolver)
        url = [a for a in self.poly.actions if a.name == action][0].url
        response = None

        if self.alps is not None:
            descriptor = self.alps.get_descriptor(action)
            if descriptor is not None:
                if isinstance(descriptor, Unsafe):
                    headers = {'Accept': 'application/json'}
                    response = requests.post(url, headers=headers, json=params)
        if response is None:
            headers = {'Accept': self._accepted_types}
            response = requests.get(url, headers=headers, params=params)

        return self.formatter.parse(response.json())

    def do(self, action: str, resolver: callable = None) -> Poly:
        self.poly = self.get(action, resolver)
        self._detect_and_set_alps()
        return self.poly 

    @property 
    def alps(self) -> Alps | None:
        return self._alps

    @property
    def actions(self) -> list:
        return [a.name for a in self.poly.actions]

    def can_do(self, action: str) -> bool: 
        return 0 < len([a for a in self.poly.actions if a.name == action])

    def _detect_and_set_alps(self):
        if self.can_do('alps'):
            poly = self.get('alps')
            alps_data = first_or_default(poly.data,{})
            self._alps = Alps.parse(alps_data, Alps.default_resolver(alps_data))

    def _resolve_params(self, action: str, resolver: callable) -> dict:
        if resolver is None or self.alps is None:
            return {}
        
        descriptor = self.alps.get_descriptor(action)
        
        params = {}
        for d in descriptor.descriptors:
            params[d.id] = resolver(d)

        return params



from .poly import *

class Hal: 
    MIME_TYPE = 'application/vnd.hal+json'
    def format(poly: Poly) -> dict: 
        return {}

    def parse(hal: dict) -> Poly: 
        return Poly().set_title('HAL is currently not supported')